# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dag.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tdag.proto\x12\x08jobs.dag\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9a\x01\n\x03\x44\x41G\x12+\n\x07symbols\x18\x01 \x03(\x0b\x32\x1a.jobs.dag.DAG.SymbolsEntry\x12&\n\tflowchart\x18\x02 \x01(\x0b\x32\x13.jobs.dag.Flowchart\x1a>\n\x0cSymbolsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.jobs.dag.Node:\x02\x38\x01\"\xe5\x01\n\tFlowchart\x12-\n\x05nodes\x18\x01 \x03(\x0b\x32\x1e.jobs.dag.Flowchart.NodesEntry\x12-\n\x05\x65\x64ges\x18\x02 \x03(\x0b\x32\x1e.jobs.dag.Flowchart.EdgesEntry\x1a<\n\nNodesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.jobs.dag.Node:\x02\x38\x01\x1a<\n\nEdgesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.jobs.dag.Edge:\x02\x38\x01\"\xa8\x01\n\x04\x45\x64ge\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06source\x18\x02 \x01(\t\x12\x0e\n\x06target\x18\x03 \x01(\t\x12%\n\x04type\x18\x04 \x01(\x0e\x32\x17.jobs.dag.Edge.EdgeType\x12\x13\n\x0bnbox_string\x18\x05 \x01(\t\"8\n\x08\x45\x64geType\x12\x13\n\x0f\x45XECUTION_ORDER\x10\x00\x12\r\n\tBRANCHING\x10\x01\x12\x08\n\x04LOOP\x10\x02\"\xa2\x02\n\tRunStatus\x12)\n\x05start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x06inputs\x18\x03 \x03(\x0b\x32\x1f.jobs.dag.RunStatus.InputsEntry\x12\x31\n\x07outputs\x18\x04 \x03(\x0b\x32 .jobs.dag.RunStatus.OutputsEntry\x1a-\n\x0bInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a.\n\x0cOutputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe4\x02\n\x04\x43ode\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x13\n\x0bnbox_string\x18\x03 \x01(\t\x12\x0e\n\x06lineno\x18\x04 \x01(\x05\x12\x12\n\ncol_offset\x18\x05 \x01(\x05\x12\x12\n\nend_lineno\x18\x06 \x01(\x05\x12\x16\n\x0e\x65nd_col_offset\x18\x07 \x01(\x05\x12*\n\x06inputs\x18\x08 \x03(\x0b\x32\x1a.jobs.dag.Code.InputsEntry\x12,\n\x07outputs\x18\t \x03(\x0b\x32\x1b.jobs.dag.Code.OutputsEntry\x12\"\n\nconditions\x18\n \x03(\x0b\x32\x0e.jobs.dag.Code\x1a-\n\x0bInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a.\n\x0cOutputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x91\x02\n\x04Node\x12\n\n\x02id\x18\x01 \x01(\t\x12\x17\n\x0f\x65xecution_index\x18\x02 \x01(\x05\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x10\n\x08operator\x18\x04 \x01(\t\x12%\n\x04type\x18\x05 \x01(\x0e\x32\x17.jobs.dag.Node.NodeType\x12\x1c\n\x04info\x18\x06 \x01(\x0b\x32\x0e.jobs.dag.Code\x12\x13\n\x0bnbox_string\x18\x07 \x01(\t\x12\'\n\nrun_status\x18\x08 \x01(\x0b\x32\x13.jobs.dag.RunStatus\"A\n\x08NodeType\x12\x08\n\x04\x43ODE\x10\x00\x12\x06\n\x02OP\x10\x01\x12\x08\n\x04LOOP\x10\x02\x12\r\n\tBRANCHING\x10\x03\x12\n\n\x06RETURN\x10\x04\x42\x32Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pbb\x06proto3')



_DAG = DESCRIPTOR.message_types_by_name['DAG']
_DAG_SYMBOLSENTRY = _DAG.nested_types_by_name['SymbolsEntry']
_FLOWCHART = DESCRIPTOR.message_types_by_name['Flowchart']
_FLOWCHART_NODESENTRY = _FLOWCHART.nested_types_by_name['NodesEntry']
_FLOWCHART_EDGESENTRY = _FLOWCHART.nested_types_by_name['EdgesEntry']
_EDGE = DESCRIPTOR.message_types_by_name['Edge']
_RUNSTATUS = DESCRIPTOR.message_types_by_name['RunStatus']
_RUNSTATUS_INPUTSENTRY = _RUNSTATUS.nested_types_by_name['InputsEntry']
_RUNSTATUS_OUTPUTSENTRY = _RUNSTATUS.nested_types_by_name['OutputsEntry']
_CODE = DESCRIPTOR.message_types_by_name['Code']
_CODE_INPUTSENTRY = _CODE.nested_types_by_name['InputsEntry']
_CODE_OUTPUTSENTRY = _CODE.nested_types_by_name['OutputsEntry']
_NODE = DESCRIPTOR.message_types_by_name['Node']
_EDGE_EDGETYPE = _EDGE.enum_types_by_name['EdgeType']
_NODE_NODETYPE = _NODE.enum_types_by_name['NodeType']
DAG = _reflection.GeneratedProtocolMessageType('DAG', (_message.Message,), {

  'SymbolsEntry' : _reflection.GeneratedProtocolMessageType('SymbolsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DAG_SYMBOLSENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.DAG.SymbolsEntry)
    })
  ,
  'DESCRIPTOR' : _DAG,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.DAG)
  })
_sym_db.RegisterMessage(DAG)
_sym_db.RegisterMessage(DAG.SymbolsEntry)

Flowchart = _reflection.GeneratedProtocolMessageType('Flowchart', (_message.Message,), {

  'NodesEntry' : _reflection.GeneratedProtocolMessageType('NodesEntry', (_message.Message,), {
    'DESCRIPTOR' : _FLOWCHART_NODESENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.Flowchart.NodesEntry)
    })
  ,

  'EdgesEntry' : _reflection.GeneratedProtocolMessageType('EdgesEntry', (_message.Message,), {
    'DESCRIPTOR' : _FLOWCHART_EDGESENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.Flowchart.EdgesEntry)
    })
  ,
  'DESCRIPTOR' : _FLOWCHART,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.Flowchart)
  })
_sym_db.RegisterMessage(Flowchart)
_sym_db.RegisterMessage(Flowchart.NodesEntry)
_sym_db.RegisterMessage(Flowchart.EdgesEntry)

Edge = _reflection.GeneratedProtocolMessageType('Edge', (_message.Message,), {
  'DESCRIPTOR' : _EDGE,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.Edge)
  })
_sym_db.RegisterMessage(Edge)

RunStatus = _reflection.GeneratedProtocolMessageType('RunStatus', (_message.Message,), {

  'InputsEntry' : _reflection.GeneratedProtocolMessageType('InputsEntry', (_message.Message,), {
    'DESCRIPTOR' : _RUNSTATUS_INPUTSENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.RunStatus.InputsEntry)
    })
  ,

  'OutputsEntry' : _reflection.GeneratedProtocolMessageType('OutputsEntry', (_message.Message,), {
    'DESCRIPTOR' : _RUNSTATUS_OUTPUTSENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.RunStatus.OutputsEntry)
    })
  ,
  'DESCRIPTOR' : _RUNSTATUS,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.RunStatus)
  })
_sym_db.RegisterMessage(RunStatus)
_sym_db.RegisterMessage(RunStatus.InputsEntry)
_sym_db.RegisterMessage(RunStatus.OutputsEntry)

Code = _reflection.GeneratedProtocolMessageType('Code', (_message.Message,), {

  'InputsEntry' : _reflection.GeneratedProtocolMessageType('InputsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CODE_INPUTSENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.Code.InputsEntry)
    })
  ,

  'OutputsEntry' : _reflection.GeneratedProtocolMessageType('OutputsEntry', (_message.Message,), {
    'DESCRIPTOR' : _CODE_OUTPUTSENTRY,
    '__module__' : 'dag_pb2'
    # @@protoc_insertion_point(class_scope:jobs.dag.Code.OutputsEntry)
    })
  ,
  'DESCRIPTOR' : _CODE,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.Code)
  })
_sym_db.RegisterMessage(Code)
_sym_db.RegisterMessage(Code.InputsEntry)
_sym_db.RegisterMessage(Code.OutputsEntry)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), {
  'DESCRIPTOR' : _NODE,
  '__module__' : 'dag_pb2'
  # @@protoc_insertion_point(class_scope:jobs.dag.Node)
  })
_sym_db.RegisterMessage(Node)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z0github.com/NimbleBoxAI/jobs-architecture/jobs_pb'
  _DAG_SYMBOLSENTRY._options = None
  _DAG_SYMBOLSENTRY._serialized_options = b'8\001'
  _FLOWCHART_NODESENTRY._options = None
  _FLOWCHART_NODESENTRY._serialized_options = b'8\001'
  _FLOWCHART_EDGESENTRY._options = None
  _FLOWCHART_EDGESENTRY._serialized_options = b'8\001'
  _RUNSTATUS_INPUTSENTRY._options = None
  _RUNSTATUS_INPUTSENTRY._serialized_options = b'8\001'
  _RUNSTATUS_OUTPUTSENTRY._options = None
  _RUNSTATUS_OUTPUTSENTRY._serialized_options = b'8\001'
  _CODE_INPUTSENTRY._options = None
  _CODE_INPUTSENTRY._serialized_options = b'8\001'
  _CODE_OUTPUTSENTRY._options = None
  _CODE_OUTPUTSENTRY._serialized_options = b'8\001'
  _DAG._serialized_start=57
  _DAG._serialized_end=211
  _DAG_SYMBOLSENTRY._serialized_start=149
  _DAG_SYMBOLSENTRY._serialized_end=211
  _FLOWCHART._serialized_start=214
  _FLOWCHART._serialized_end=443
  _FLOWCHART_NODESENTRY._serialized_start=321
  _FLOWCHART_NODESENTRY._serialized_end=381
  _FLOWCHART_EDGESENTRY._serialized_start=383
  _FLOWCHART_EDGESENTRY._serialized_end=443
  _EDGE._serialized_start=446
  _EDGE._serialized_end=614
  _EDGE_EDGETYPE._serialized_start=558
  _EDGE_EDGETYPE._serialized_end=614
  _RUNSTATUS._serialized_start=617
  _RUNSTATUS._serialized_end=907
  _RUNSTATUS_INPUTSENTRY._serialized_start=814
  _RUNSTATUS_INPUTSENTRY._serialized_end=859
  _RUNSTATUS_OUTPUTSENTRY._serialized_start=861
  _RUNSTATUS_OUTPUTSENTRY._serialized_end=907
  _CODE._serialized_start=910
  _CODE._serialized_end=1266
  _CODE_INPUTSENTRY._serialized_start=814
  _CODE_INPUTSENTRY._serialized_end=859
  _CODE_OUTPUTSENTRY._serialized_start=861
  _CODE_OUTPUTSENTRY._serialized_end=907
  _NODE._serialized_start=1269
  _NODE._serialized_end=1542
  _NODE_NODETYPE._serialized_start=1477
  _NODE_NODETYPE._serialized_end=1542
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbox',
 'nbox.framework',
 'nbox.hyperloop',
 'nbox.lib',
 'nbox.nbxlib',
 'nbox.sub_utils']

package_data = \
{'': ['*'], 'nbox': ['assets/*'], 'nbox.framework': ['protos/*']}

install_requires = \
['Jinja2==3.0.3',
 'dill==0.3.4',
 'grpcio==1.43.0',
 'mypy-protobuf==3.2.0',
 'protobuf==4.21.1',
 'python-json-logger==2.0.2',
 'randomname>=0.1.3,<0.2.0',
 'requests>=2.25.1,<3.0.0',
 'tabulate==0.8.9']

setup_kwargs = {
    'name': 'nbox',
    'version': '0.9.14rc11',
    'description': 'ML Inference 🥶',
    'long_description': '<a href="https://nimblebox.ai/" target="_blank"><img src="./assets/built_at_nbx.svg" align="right"></a>\n\n# 🏖️ Nbox\n\n`nbox` is NimbleBox.ai\'s official SDK.\n\n> The entire purpose of this package is to make using ML 🥶.\n\n```\npip install nbox\n```\n\n## 🔥 Usage\n\n`nbox` provides first class support API for all NimbleBox.ai infrastructure (NBX-Build, Jobs, Deploy) and services (NBX-Workspaces) components. Write jobs using `nbox.Operators`:\n\n```python\nfrom nbox import Operator\nfrom nbox.nbxlib.ops import Magic\n\n# define a class object\nweekly_trainer: Operator = Magic()\n\n# call your operators\nweekly_trainer(\n  pass_values = "directly",\n)\n\n# confident? deploy it to your cloud\nweekly_trainer.deploy(\n  job_id_or_name = "magic_jobs",\n  schedule = Schedule(4, 30, [\'fri\']) # schedule like humans\n)\n```\n\nDeploy your machine learning or statistical models:\n\n```python\nfrom nbox import Model\nfrom transformers import AutoModelForSequenceClassification, AutoTokenizer\n\n# define your pre and post processing functions\ndef pre(x: Dict):\n  return AutoTokenizer(**x)\n\n# load your classifier with functions\nmodel = AutoModelForSequenceClassification.from_pretrained("distill-bert")\nclassifier = Model(model, pre = pre)\n\n# call your model\nclassifier(f"Is this a good picture?")\n\n# get full control on exporting it\nspec = classifier.torch_to_onnx(\n  TorchToOnnx(...)\n)\n\n# confident? deploy it your cloud\nurl, key = classifier.deploy(\n  spec, deployment_id_or_name = "classification"\n)\n\n# use it anywhere\npred = requests.post(\n  url,\n  json = {\n    "text": f"Is this a good picture?"\n  },\n  header = {"Authorization": f"Bearer {key}"}\n).json()\n```\n\n# 🧩 License\n\nThe code in thist repo is licensed as [Apache License 2.0](./LICENSE). Please check for individual repositories for licenses.\n',
    'author': 'NBX Research',
    'author_email': 'research@nimblebox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NimbleBoxAI/nbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# weird file name, yessir. Why? because
# from nbox.framework.on_functions import PureFunctionParser
# read as "from nbox's framework on Functions import the pure-function Parser"

import ast
import base64
import inspect
from typing import Union
from uuid import uuid4

from ..utils import logger
from ..hyperloop.dag_pb2 import DAG, Flowchart, Node, Code, Edge, RunStatus

# ==================


# ==================
# the next set of functions are meant as support methods to create nbox_strings IR.

class NboxStrings:
  OP_TO_STRING = {
    "function": "FUNCTION: {name} ( {inputs} ) => [ {outputs} ]",
    "define": "DEFINE: {name} ( {inputs} )",
    "for": "FOR: ( {iter} ) => [ {target} ]",
    "return": "RETURN: [ {value} ]",
  }

  def __init__(self):
    pass

  def function(self, name, inputs, outputs):
    return self.OP_TO_STRING["function"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs]),
      outputs=", ".join(outputs)
    )

  def define(self, name, inputs):
    return self.OP_TO_STRING["define"].format(
      name=name,
      inputs=", ".join([f"{x['kwarg']}={x['value']}" for x in inputs])
    )

  def for_loop(self, iter, target):
    return self.OP_TO_STRING["for"].format(
      iter=iter,
      target=", ".join(target)
    )

  def return_statement(self, value):
    return self.OP_TO_STRING["return"].format(
      value=value
    )

nbxl = NboxStrings()

def write_program(nodes):
  for i, n in enumerate(nodes):
    logger.debug(f"{i:03d}|{n.get('nbox_string', n.get('info').get('nbox_string'))}")


# ==================

def get_code_portion(cl, lineno, col_offset, end_lineno, end_col_offset, b64 = True, **_):
  sl, so, el, eo = lineno, col_offset, end_lineno, end_col_offset
  if sl == el:
    return cl[sl-1][so:eo]
  code = ""
  for i in range(sl - 1, el, 1):
    if i == sl - 1:
      code += cl[i][so:]
    elif i == el - 1:
      code += "\n" + cl[i][:eo]
    else:
      code += "\n" + cl[i]
  
  # convert to base64
  if b64:
    return base64.b64encode(code.encode()).decode()
  return code

def parse_args(node):
  inputs = []
  for a in node.args:
    a = a.arg if isinstance(a, ast.arg) else a
    inputs.append({
      "kwarg": None,
      "value": a,
    })
  for a in node.kwonlyargs:
    inputs.append({
      "kwarg": a[0],
      "value": a[1],
    })
  if node.vararg:
    inputs.append({
      "kwarg": "*"+node.vararg.arg,
      "value": None
    })
  if node.kwarg:
    inputs.append({
      "kwarg": "**"+node.kwarg.arg,
      "value": None
    })
  return inputs

def get_name(node):
  if isinstance(node, ast.Name):
    return node.id
  elif isinstance(node, ast.Constant):
    return node.value
  elif isinstance(node, ast.Attribute):
    return get_name(node.value) + "." + node.attr
  elif isinstance(node, ast.Call):
    return get_name(node.func)

def parse_kwargs(node, lines):
  if isinstance(node, ast.Name):
    return node.id
  elif isinstance(node, ast.Constant):
    val = node.value
    return val
  elif isinstance(node, ast.keyword):
    arg = node.arg
    value = node.value
    if 'id' in value.__dict__:
      # arg = my_model
      return (arg, value.id)
    elif 'value' in value.__dict__:
      if isinstance(value.value, ast.Call):
        # arg = my_model(...)
        return (arg, get_code_portion(lines, b64 = False, **value.value.__dict__))
      else:
        # arg = 20
        return (arg, value.value)
    elif 'func' in value.__dict__:
      #   arg = some_function(with, some=args)
      #   ^^^   ^^^^^
      # kwarg   value
      return {"kwarg": arg, "value": get_code_portion(lines, b64 = False, **value.__dict__)}
  elif isinstance(node, ast.Call):
    return get_code_portion(lines, **node.func.__dict__)

def node_assign_or_expr(node, lines, node_proto: Node) -> Union[Node, None]:
  # print(get_code_portion(lines, b64 = False, **node.__dict__))
  value = node.value
  try:
    name = get_name(value.func)
  except AttributeError:
    return None
  args = [parse_kwargs(x, lines) for x in value.args + value.keywords]
  inputs = []
  for a in args:
    if isinstance(a, dict):
      inputs.append(a)
      continue
    inputs.append({
      "kwarg": a[0] if isinstance(a, tuple) else None,
      "value": a[1] if isinstance(a, tuple) else a,
    })

  outputs = []
  if isinstance(node, ast.Assign):
    targets = node.targets[0]
    outputs = [parse_kwargs(x, lines) for x in targets.elts] \
      if isinstance(targets, ast.Tuple) \
      else [parse_kwargs(targets, lines)
    ]

  nbox_string = nbxl.function(name, inputs, outputs)

  node_proto.MergeFrom(Node(
    name = name,
    type = Node.NodeType.OP,
    info = Code(
      name = name,
      nbox_string = nbox_string,
      code = get_code_portion(lines, **node.__dict__),
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {str(x["kwarg"]): str(x["value"]) for x in inputs},
      outputs = {x: "" for x in outputs},
    )
  ))
  return node_proto

def node_if_expr(node: ast.IfExp, lines: list, node_proto: Node) -> Node:
  def get_conditions(node, lines, conds = []):
    if not hasattr(node, "test"):
      else_cond = list(filter(lambda x: x["condition"] == "else", conds))
      if not else_cond:
        conds.append({
          "condition": "else",
          "code": dict(
            lineno = node.lineno,
            col_offset = node.col_offset,
            end_lineno = node.end_lineno,
            end_col_offset = node.end_col_offset,
          )
        })
      else:
        cond = else_cond[0]
        cond["code"]["end_lineno"] = node.end_lineno
        cond["code"]["end_col_offset"] = node.end_col_offset
    else:
      condition = get_code_portion(lines, **node.test.__dict__)

      # need to run this last or "else" comes up first
      conds.append({
        "condition": condition,
        "code": {
          "lineno": node.lineno,
          "col_offset": node.col_offset,
          "end_lineno": node.end_lineno,
          "end_col_offset": node.end_col_offset,
        }
      })
      for x in node.orelse:
        get_conditions(x, lines, conds)

    return conds

  all_conditions = get_conditions(node, lines, conds = [])
  ends = []
  for b0, b1  in zip(all_conditions[:-1], all_conditions[1:]):
    ends.append([b0["code"], b1["code"]])
  for i in range(len(ends)):
    ends[i] = {
      "lineno": ends[i][0]["lineno"],
      "col_offset": ends[i][0]["col_offset"],
      "end_lineno": ends[i][1]["lineno"],
      "end_col_offset": ends[i][1]["col_offset"],
    }
  ends += [all_conditions[-1]["code"]]

  conditions = []
  for i, c in enumerate(all_conditions):
    box = ends[i]
    _node = Code(
      name = f"if-{i}",
      nbox_string = c["condition"],
      lineno = box['lineno'],
      col_offset = box['col_offset'],
      end_lineno = box['end_lineno'],
      end_col_offset = box['end_col_offset'],
      inputs = [],
      outputs = [],
    )
    conditions.append(_node)

  # update the node_proto
  node_proto.MergeFrom(Node(
    name = "if",
    type = Node.NodeType.BRANCHING,
    info = Code(
      name = "if",
      nbox_string = "IF: { " + ", ".join(x.nbox_string for x in conditions) + " }",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {},
      conditions = conditions,
    )
  ))
  return node_proto

def node_for_expr(node: ast.For, lines: list, node_proto: Node) -> Node:
  iter_str = get_code_portion(lines, **node.iter.__dict__)
  if isinstance(node.target, ast.Tuple):
    targets = [x.id for x in node.target.elts]
  else:
    targets = [get_name(node.target)]
  code_ = get_code_portion(lines, **node.__dict__)
  nbox_string = nbxl.for_loop(iter_str, targets)

  # update node_proto
  node_proto.MergeFrom(Node(
    name = "for",
    type = Node.NodeType.LOOP,
    info = Code(
      name = "for",
      code = code_,
      nbox_string = nbox_string,
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':x for x in targets},
    )
  ))
  return node_proto

def node_return(node: ast.Return, lines, node_proto: Node) -> Node:
  if isinstance(node.value, ast.Tuple):
    returns = [x.id for x in node.value.elts]
  else:
    returns = [get_name(node.value)]
  node_proto.MergeFrom(Node(
    name = "return",
    # type = Node.NodeType.RETURN,
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = nbxl.return_statement(returns),
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto

def def_func_or_class(node, lines, node_proto: Node) -> Node:
  print(node)

  node_proto.MergeFrom(Node(
    name = f"symbol-{node.name}",
    type = Node.NodeType.OP, # TODO: @yashbonde expand node types to include class/func definitions
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = "",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto

def async_def_func_or_class(node, lines, node_proto: Node) -> Node:
  print(node.__dict__)
  node_proto.MergeFrom(Node(
    name = f"symbol-{node.name}",
    type = Node.NodeType.OP, # TODO: @yashbonde expand node types to include class/func definitions
    info = Code(
      name = "return",
      code = get_code_portion(lines, **node.__dict__),
      nbox_string = "",
      lineno = node.lineno,
      col_offset = node.col_offset,
      end_lineno = node.end_lineno,
      end_col_offset = node.end_col_offset,
      inputs = {},
      outputs = {'None':'None'},
    )
  ))
  return node_proto


# ==================

type_wise_logic = {
  # defns ------
  ast.FunctionDef: def_func_or_class,
  ast.ClassDef: def_func_or_class,
  ast.AsyncFunctionDef: def_func_or_class,

  # nodes ------
  ast.Assign: node_assign_or_expr,
  ast.Expr: node_assign_or_expr,
  ast.If: node_if_expr,
  ast.For: node_for_expr,

  # Return ------
  ast.Return: node_return,

  # todos ------
  # ast.AsyncFunctionDef: async_func_def,
  # ast.Await: node_assign_or_expr,
}

# ==================


def code_node(execution_index, expr, code_lines) -> Node:
  # code pieces that are not yet supported should still see the code
  return Node(
    id = str(uuid4()),
    execution_index = execution_index,
    name = f"codeblock-{execution_index}",
    type = Node.NodeType.OP,
    operator = "CodeBlock",
    nbox_string = f"CODE: {str(type(expr))}", # :str
    run_status = RunStatus(), # no need to initialise this object
    info = Code(
      name = f"codeblock-{execution_index}", # :str
      code = get_code_portion(code_lines, bs64 = True, **expr.__dict__), # :str (base64)
      nbox_string = None, # :str
      lineno = expr.lineno, # :int
      col_offset = expr.col_offset, # :int
      end_lineno = expr.end_lineno, # :int
      end_col_offset = expr.end_col_offset, # :int
      inputs = [], # :list[Dict[Any, Any]]
      outputs = [], # :list[str]
    )
  )


def get_nbx_flow(forward) -> DAG:
  """Get NBX flowchart. Read python grammar here:
  https://docs.python.org/3/reference/grammar.html

  Args:
      forward (callable): the function whose flowchart is to be generated
  """
  if type(forward) == str:
    code = forward
  else:
    code = inspect.getsource(forward).strip() # get code string from operator
  code_lines = code.splitlines()
  ast_node = ast.parse(code)

  edges = {} # this is the flow
  nodes = {} # this is the operators
  symbols_to_nodes = {} # this is things that are defined at runtime

  try:
    iter_obj = ast_node.body[0].body if hasattr(ast_node.body[0], 'body') else ast_node.body
    for i, expr in enumerate(iter_obj):
      # create the empty node that will be used everywhere
      if not type(expr) in type_wise_logic:
        node: Node = code_node(i, expr, code_lines)
        nodes[node.id] = node
        continue

      # define an initial proto and then then other functions will fill it up
      node_proto = Node(id = str(uuid4()), execution_index = i)
      output = type_wise_logic[type(expr)](expr, code_lines, node_proto)
      if output is None:
        continue
      nodes[node_proto.id] = node_proto

    # edges for execution order can be added
    _node_ids = tuple(nodes.keys())
    for op0, op1 in zip(_node_ids[:-1], _node_ids[1:]):
      _id = f"edge-{op0}-X-{op1}"
      edges[_id] = Edge(
        id = _id,
        source = op0,
        target = op1,
        type = Edge.EdgeType.EXECUTION_ORDER,
        nbox_string = None
      )
  except Exception as e:
    logger.error("Error in parsing the job-flow")
    logger.error(e)
    nodes = {}
    edges = {}
    symbols_to_nodes = {}

  return DAG(flowchart = Flowchart(nodes = nodes, edges = edges), symbols=symbols_to_nodes)



"""
I started with the initial assumption that abstracting away the details of
processing the data was doable but unfortunately I was not able to get it to
work reliably. We redid the entire thing by manually defining the code and
operations but it was still super hard. So the answer was to generate code
that the user will have to implement.

It loosely works like this, let's just say I am trying to write code for
``torch_to_onnx`` module. ``ModelSpec`` expects that ``source`` and ``target``
contain all the required information for running this model in another
location. Now the actual code is only of two lines, but the many different
frameworks, with their different functions and different inputs means that
any scope of manual process is gone out the window. So we basically create a
registry that can store all the information and pass it along to ``nbox.Model``.

The implementation of the registry is a ``@register`` decorator, which takes
in all the user information and underneath structures the entire thing and
stores it, if needed to be written down for compilation.

Read the code for best understanding.
"""

import os
from typing import Callable

from nbox import utils as U
from nbox.utils import logger
from nbox.framework.autogen import ml_register
from nbox.framework.model_spec_pb2 import ModelSpec, Tensor

class NboxOptions:
  def __init__(self, model_name: str = None, folder: str = None, create_folder = False):
    model_name = model_name or U.get_random_name().replace("-", "_")
    if create_folder:
      self.folder = U.join(folder if folder != None else U.NBOX_HOME_DIR, model_name)
      self.model_name = model_name
      os.makedirs(self.folder, exist_ok = False)
    else:
      self.folder = folder
      self.model_name = model_name


# """
# ## Torch

# Torch is a deep learning framework from Facebook. It is aimed towards a simple interface and is a huge inspiration
# to me. User needs to define the config/init and forward pass. This is the minimum input you need to take from
# a user to get all the output.
# """

# def torch_condition(model) -> bool:
#   import torch
#   if isinstance(model, torch.nn.Module):
#     return True
#   return False

# ml_register.register_package(
#   package_name = "torch",
#   conditional_fn = torch_condition,
#   requirements_txt = ["-f https://download.pytorch.org/whl/cpu/torch_stable.html", "torch"],
# )


# @ml_register.torch(
#   _from = "torchscript", # from_torchscript
#   stub_function_import_string = "from torch.jit import load",
#   ignore_args = ["f"],
# )
# def load_torchscript(user_options: 'FromTorchScript', nbox_options: NboxOptions, model_spec: ModelSpec) -> 'torch.nn.Module':
#   from torch.jit import load
#   model = load(U.join(nbox_options.folder, "model.pt"), map_location="cpu")
#   return model


# @ml_register.torch(
#   _to = "torchscript", # to_torchscript
#   stub_function_import_string = "from torch.jit import trace",
#   ignore_args = ["filename"],
# )
# def torch_to_torchcript(user_options, nbox_options: NboxOptions, model_spec: ModelSpec) -> ModelSpec:
#   import torch
#   from torch.jit import trace
#   traced_model = trace(**user_options.__dict__)
#   filepath = U.join(nbox_options.folder, "model.pt")
#   model_spec.source.path = filepath
#   model_spec.inputs.extend([
#     Tensor(
#       name = str(i),
#       shape = tuple(t.shape),
#       dtype = str(t.numpy().dtype)
#     ) for i, t in enumerate(user_options.example_inputs)
#   ])

#   # actual export
#   torch.jit.save(traced_model, filepath)
#   return model_spec


# @ml_register.torch(
#   _to = "onnx",
#   stub_function_import_string = "from torch.onnx import export",
#   ignore_args = ["f", "model"],
# )
# def torch_to_onnx(user_options, nbox_options, spec: ModelSpec) -> ModelSpec:
#   import torch
#   from torch.onnx import export
#   user_options.f = nbox_options.filepath # override the folder
#   inputs = user_options.args
#   if isinstance(inputs, torch.Tensor):
#     spec.inputs.append(Tensor(name = '0', shape = tuple(inputs.shape), dtype = str(inputs.numpy().dtype)))
#   elif isinstance(inputs, tuple):
#     if any(isinstance(x, dict) for x in inputs):
#       if not inputs[-1] == {}:
#         logger.warning(f"Last input is not an empty dictionary and is recommended by ONNX Export")
#     for i,t in enumerate(inputs):
#       if isinstance(t, torch.Tensor):
#         spec.inputs.append(Tensor(name = str(i), shape = tuple(t.shape), dtype = str(t.numpy().dtype)))
#       elif isinstance(t, dict):
#         for k,v in t.items():
#           spec.inputs.append(Tensor(name = k, shape = tuple(v.shape), dtype = str(v.numpy().dtype)))

#   # actual export
#   export(**user_options.__dict__)
#   return spec


# """
# ## ONNX

# OnnxRuntime is a C++ library that allows you to run ONNX models in your application. It is a wrapper around
# the ONNX API. It *does not* work on Apple M1 machines and so suck it up! But this is useful for server
# side processing.
# """

# def _check_onnx_model(model):
#   import onnx
#   if isinstance(model, onnx.onnx_ml_pb2.ModelProto):
#     return True
#   return False

# ml_register.register_package(
#   package_name = "onnx_runtime",
#   conditional_fn = _check_onnx_model,
#   requirements_txt = ["onnxruntime", "numpy", "onnx"]
# )

# @ml_register.onnx(
#   _from = "onnx",
# )
# def load_onnx(user_options: 'FromOnnx', nbox_options: NboxOptions, model_spec: ModelSpec) -> Callable:
#   import onnxruntime as ort
#   sess = ort.InferenceSession(U.join(nbox_options.folder, "model.onnx"))
#   return sess._run


# class ONNXRtModel(FrameworkAgnosticProtocol):
#   @U.isthere("onnx", "onnxruntime", "numpy", soft = False)
#   def __init__(self, m0):

#   def forward(self, input_object) -> ModelOutput:
#     # import numpy as np
#     # model_inputs = self._logic(input_object)
#     # input_names = [i.name for i in self._sess.get_inputs()]
#     # output_names = [o.name for o in self._sess.get_outputs()]
#     # if isinstance(model_inputs,dict):
#     #   tensor_dict = model_inputs
#     #   if isinstance(list(model_inputs.values())[0],SklearnInput):
#     #     tensor_dict = {list(model_inputs.keys())[0]:list(model_inputs.values())[0].inputs}
#     #   model_inputs = {name: tensor for name, tensor in zip(input_names, tuple(tensor_dict.values()))}
#     # else:
#     #   raise ValueError("Your Logic needs to return dictionary")

#     # # conversion to proper arrays -> if can be numpy-ed and float32
#     # for key, value in model_inputs.items():
#     #   if hasattr(value, "numpy"):
#     #     value = value.numpy()
#     #   if value.dtype in (np.float64, np.float32, np.float16):
#     #     value = value.astype(np.float32)
#     #   elif value.dtype in (np.int64, np.int32, np.int16, np.int8, np.int64):
#     #     value = value.astype(np.int32)
#     #   model_inputs[key] = value

#     # res = self._sess.run(output_names=output_names, input_feed=model_inputs)

#     res = self._sess.run(**input_object)
#     return ModelOutput(
#       inputs = input_object,
#       outputs = res,
#     )

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Any:
#     import onnx
#     logger.debug(f"Deserialising ONNX model from {model_meta.export_path}")

#     if model_meta.export_type == "onnx":
#       model = onnx.load(model_meta.load_kwargs["model"])
#       return model
#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")




# """
# ## Sklearn

# Scikit-learn is a Python library that allows you to use classical machine
# learning algorithms in your application. Its versatility is huge and that
# means that a dedicate Input DataModel is required.
# """

# def _check_sklearn_model(model) -> bool:
#   if "sklearn" in type(model):
#     return True
#   return False

# ml_register.register_package(
#   package_name = "sklearn",
#   conditional_fn = _check_sklearn_model,
#   requirements_txt = "scikit-learn",
# )

# @ml_register.register_deserializer(
#   package_name = "sklearn",
#   framework_name = "joblib", # from_joblib
#   stub_name = "FromJoblib", 
#   stub_function_import_string = "from joblib import load",
#   ignore_args = [
#     "filename"
#   ],
# )
# def sklearn_from_joblib(user_options, nbox_options: NboxOptions, model_spec: ModelSpec):
#   import joblib
#   model = joblib.load(nbox_options.filepath)
#   return model

# @ml_register.register_serializer(
#   package_name = "sklearn",
#   framework_name = "joblib", # to_joblib
#   stub_name = "ToJoblib",
#   stub_function_import_string = "from joblib import dump",
#   ignore_args = [
#     "filename"
#   ],
# )
# def sklearn_to_joblib(user_options, nbox_options: NboxOptions, model_spec: ModelSpec):
#   from joblib import dump
#   dump(user_options, nbox_options.filepath)
#   return model_spec

# class SklearnModel(FrameworkAgnosticProtocol):
#   @U.isthere("sklearn", "numpy", soft = False)
#   def __init__(self, m1):
#     logger.debug(f"Trying to load as Sklearn")
#     if not "sklearn" in str(type(m1)):
#       raise InvalidProtocolError

#     self._model = m1

#     self.exporters = {
#       "pkl": U.to_pickle
#     }
#     self.loaders = {
#       "pkl": U.from_pickle
#     }

#   def forward(self, input_object: Dict) -> ModelOutput:
#     """Sklearn covers a far wider range of models and is therefore more complicated to cover.
#     For now the input is expected to be a Dict that has "inputs" dict that will be fed to the model.
#     And it has "method" key that will be called upon the model i.e. whether user wants
#     ``predict`` (default) or ``predict_proba`` is left upto the user.
#     """

#     # if "sklearn.neighbors.NearestNeighbors" in str(type(self._model)):
#     #   method = getattr(self._model, "kneighbors") if method == None else getattr(self._model, method)
#     # elif "sklearn.cluster" in str(type(self._model)):
#     #   if any(
#     #     x in str(type(self._model)) for x in ["AgglomerativeClustering", "DBSCAN", "OPTICS", "SpectralClustering"]
#     #   ):
#     #     method = getattr(self._model, "fit_predict")
#     # else:
#     #   try:
#     #     method = getattr(self._model, "predict") if method == None else getattr(self._model, method)
#     #   except Exception as e:
#     #     logger.debug(f"[ERROR] Model Prediction Function is not yet registered {e}")

#     model_input = input_object.get("inputs", None)
#     method = input_object.get("method", "predict")
#     method_fn = getattr(self._model, method, None)
#     if method_fn == None:
#       logger.error(f"Method {method} not found in SklearModel")
#       raise ValueError(f"Method {method} not found in model")
#     out = method(**model_input)

#     return ModelOutput(inputs = model_input, outputs = out,)

#   @U.isthere("skl2onnx", soft = False)
#   def export_to_onnx(self, input_object, export_model_path, model_file_name="model.onnx", opset_version=None,) -> ModelSpec:
#     import skl2onnx
#     import skl2onnx.common.data_types as dt

#     __NP_DTYPE_TO_SKL_DTYPE = {
#       "bool_": dt.BooleanTensorType,
#       "double": dt.DoubleTensorType,
#       "int16": dt.Int16TensorType,
#       "complex128": dt.Complex128TensorType,
#       "int32": dt.Int32TensorType,
#       "uint16": dt.UInt16TensorType,
#       "complex64": dt.Complex64TensorType,
#       "float16": dt.Float16TensorType,
#       "int64": dt.Int64TensorType,
#       "str_": dt.StringTensorType,
#       "uint32": dt.UInt32TensorType,
#       "float64": dt.FloatTensorType,
#       "uint64": dt.UInt64TensorType,
#       "int8": dt.Int8TensorType,
#       "uint8": dt.UInt8TensorType,
#     }

#     export_path = join(export_model_path, model_file_name)
#     method = input_object.get("method", "predict")
#     method = getattr(self._model, method, None)
#     if method == None:
#       logger.error(f"Method {method} not found in SklearModel")
#       raise ValueError(f"Method {method} not found in model")
#     iod = get_io_dict(input_object, method, self.forward)

#     # the args need to be converted to proper formatted data types
#     initial_types = []
#     input_names, input_shapes, input_dtypes = [],[],[]

#     # Case for input object with single tensor
#     inputs=iod["inputs"]
#     if isinstance(inputs['name'], str):
#       input_names.append(inputs["name"])
#       input_dtypes.append(inputs["dtype"])
#       tensorShape = list()
#       for i in inputs["tensorShape"]["dim"]:
#         tensorShape.append(i["size"])
#       input_shapes.append(tensorShape)
#     else:
#       for key in inputs:
#         input_names.append(inputs[key]['name'])
#         input_dtypes.append(inputs[key]['dtype'][key])
#         tensorShape = list()
#         for i in inputs[key]["tensorShape"]["dim"]:
#           tensorShape.append(i["size"])
#         input_shapes.append(tensorShape)

#     for name, shape, dtype in zip(input_names, input_shapes, input_dtypes):
#       shape[0] = None # batching requires the first dimension to be None
#       initial_types.append((name, __NP_DTYPE_TO_SKL_DTYPE[str(dtype)](shape)))
#     onx_model = skl2onnx.to_onnx(self._model, initial_types=initial_types, target_opset=opset_version)

#     with open(export_path, "wb") as f:
#       f.write(onx_model.SerializeToString())

#     return ModelSpec(
#       export_type="onnx",
#       export_path=export_path,
#       load_class=self.__class__.__name__,
#       load_method="from_to_onnx",
#       load_kwargs={
#         "model": f"./{model_file_name}",
#         "map_location": "cpu",
#       },
#       io_dict=iod,
#     )

#   def export_to_pkl(self, input_object, export_model_path, model_file_name = "model.pkl"):
#     export_path = join(export_model_path, model_file_name)
#     with open(export_path, "wb") as f:
#       dill.dump(self._model, f)

#     method = input_object.get("method", None)
#     method = getattr(self._model, "predict") if method == None else getattr(self._model, method)
#     iod = get_io_dict(input_object, method, self.forward)

#     return ModelSpec(
#       export_type = "pkl",
#       load_kwargs = {
#         "model": f"./{model_file_name}",
#         "map_location": "cpu",
#       },
#       io_dict = iod,
#     )

#   def export(self, format, input_object, export_model_path, **kwargs) -> ModelSpec:
#     if format == "onnx":
#       model_spec = self.export_to_onnx(input_object, export_model_path=export_model_path,**kwargs)
#     elif format == "pkl":
#       model_spec = self.export_to_pkl(input_object, export_model_path=export_model_path, **kwargs)
#     else:
#       raise InvalidProtocolError(f"Unsupported export format for torch Module: {format}")

#     import sklearn
#     model_spec.src_framework = "sklearn"
#     model_spec.src_framework_version = sklearn.__version__

#     return model_spec

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
#     logger.info(f"Deserialising SkLearn model")
#     kwargs = model_meta.load_kwargs
#     if model_meta.export_type == "pkl":
#       with open(kwargs["model"], "rb") as f:
#         model = dill.load(f)

#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")

#     return model, logic


# ################################################################################
# # Tensorflow
# # ==========
# # Tensorflow is a high-level machine learning library built by Google. It is
# # part of a much larger ecosystem called tensorflow extended.
# ################################################################################

# class TensorflowModel(FrameworkAgnosticProtocol):
#   @U.isthere("tensorflow", soft = False)
#   def __init__(self, m0):
#     import tensorflow as tf
#     logger.debug(f"Trying to load Tensorflow model")
#     if not isinstance(m0, (tf.keras.Model, tf.Module)):
#       raise InvalidProtocolError(f"First input must be a tensorflow module or Keras model, got: {type(m0)}")

#     self._model = m0

#   def forward(self, input_object: Any) -> ModelOutput:
#     model_inputs = self._logic(input_object)
#     if isinstance(model_inputs, dict):
#       out = self._model(list(model_inputs.values()))
#     else:
#       out = self._model(model_inputs)
#     return ModelOutput(inputs = input_object, outputs = out)

#   def _serialise_logic(self, fpath):
#     logger.debug(f"Saving logic to {fpath}")
#     with open(fpath, "wb") as f:
#       dill.dump(self._logic, f)

#   @U.isthere("tf2onnx", soft = False)
#   def export_to_onnx(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name="logic.dill",
#     model_file_name="model.onnx",
#     opset_version=None,
#   ) -> ModelSpec:
#     import tf2onnx
#     import tensorflow as tf

#     input_object = self._logic(input_object)

#     self._serialise_logic(join(export_model_path, logic_file_name))
#     export_path = join(export_model_path, model_file_name)
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     if "keras" in str(self._model.__class__):
#       model_proto, _ = tf2onnx.convert.from_keras(self._model, opset=opset_version)
#       with open(export_path, "wb") as f:
#         f.write(model_proto.SerializeToString())

#       return ModelSpec(
#         src_framework="tf",
#         src_framework_version=tf.__version__,
#         export_type="onnx",
#         export_path=export_path,
#         load_class=self.__class__.__name__,
#         load_method="from_convert_from_keras",
#         load_kwargs={
#           "model": f"./{model_file_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}",
#         },
#         io_dict=iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported ONNX export for tensorflow model: {type(self._model)}")

#   def export_to_savemodel(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name = "logic.dill",
#     model_dir_name = "model",
#     include_optimizer = True,
#   ) -> ModelSpec:
#     import tensorflow as tf

#     input_object = self._logic(input_object)
#     self._serialise_logic(join(export_model_path, logic_file_name))

#     export_path = join(export_model_path, model_dir_name)
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     if "keras" in str(self._model.__class__):
#       tf.keras.models.save_model(
#         self._model,
#         export_path,
#         signatures = None,
#         options = None,
#         include_optimizer = include_optimizer,
#         save_format = "tf"
#       )

#       return ModelSpec(
#         src_framework = "tf",
#         src_framework_version = tf.__version__,
#         export_type = "SaveModel",
#         export_path = export_path,
#         load_class = self.__class__.__name__,
#         load_kwargs = {
#           "model": f"./{model_dir_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}"
#         },
#         io_dict = iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported SaveModel export for tensorflow model: {type(self._model)}")

#   def export_to_h5(
#     self,
#     input_object,
#     export_model_path,
#     logic_file_name = "logic.dill",
#     model_file_name = "model.h5py",
#     include_optimizer = True,
#   ) -> ModelSpec:

#     import tensorflow as tf

#     self._serialise_logic(join(export_model_path, logic_file_name))
#     iod = get_io_dict(input_object, self._model.call, self.forward)

#     export_path = join(export_model_path, model_file_name)

#     if "keras" in str(self._model.__class__):
#       tf.keras.models.save_model(
#         self._model,
#         export_path,
#         signatures=None,
#         options=None,
#         include_optimizer=include_optimizer,
#         save_format = "h5"
#       )

#       return ModelSpec(
#         src_framework = "tf",
#         src_framework_version = tf.__version__,
#         export_type = "h5",
#         export_path = export_path,
#         load_class = self.__class__.__name__,
#         load_method = "from_savemodel",
#         load_kwargs = {
#           "model": f"./{model_file_name}",
#           "map_location": "cpu",
#           "logic_path": f"./{logic_file_name}"
#         },
#         io_dict = iod,
#       )
#     else:
#       raise InvalidProtocolError(f"Unsupported h5 export for tensorflow model: {type(self._model)}")

#   def export(self, format: str, input_object: Any, export_model_path: str, **kwargs) -> ModelSpec:
#     if format == "onnx":
#       return self.export_to_onnx(
#         input_object, export_model_path = export_model_path, **kwargs
#       )
#     elif format == "SaveModel":
#       return self.export_to_savemodel(
#         input_object, export_model_path=export_model_path, **kwargs
#       )
#     elif format == "h5":
#       return self.export_to_h5(
#         input_object, export_model_path=export_model_path, **kwargs
#       )
#     else:
#       raise  InvalidProtocolError(f"Unknown format: {format}")

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:

#     logger.info(f"Deserialising Tensorflow model from {model_meta.export_path}")
#     kwargs = model_meta.load_kwargs

#     if model_meta.export_type in ["SaveModel", "h5"]:
#       lp = kwargs.pop("logic_path")
#       logger.info(f"Loading logic from {lp}")
#       with open(lp, "rb") as f:
#         logic = dill.load(f)

#       import tensorflow as tf
#       model = tf.keras.models.load_model(model_meta.load_kwargs["model"])
#       return model, logic
#     else:
#       raise InvalidProtocolError(f"Unknown format: {model_meta.export_type}")

# ################################################################################
# # Jax
# # ===
# # Jax is built by Google for high performance ML research. Read more here:
# # https://jax.readthedocs.io/en/latest/notebooks/quickstart.html
# ################################################################################

# """ We are not implementing Jax as of now"""


# ################################################################################
# # Flax
# # ===
# # Flax is built by Google, it is a neural network library and ecosystem for JAX
# # designed for flexibility. Read more here: https://github.com/google/flax
# ################################################################################

# class FlaxModel(FrameworkAgnosticProtocol):
#   @U.isthere("flax", soft = False)
#   def __init__(self, m0):
#     import flax
#     from flax import linen as nn

#     logger.debug("Trying to load Flax model")

#     if not isinstance(m0, nn.Module):
#       raise InvalidProtocolError(f"First input must be a Flax module, got: {type(m0)}")

#     if not isinstance(m1, flax.core.frozen_dict.FrozenDict):
#       # m1 should contain the model parameters
#       raise ValueError(f"Second input for flax model must be a\
#        FrozenDict containing model parameters, got: {type(m1)}")

#     self._model = m0
#     self._params = m1

#   def forward(self, input_object: Any) -> ModelOutput:
#     model_inputs = input_object
#     if type(model_inputs) is dict:
#       out = self._model.apply(self._params, **model_inputs)
#     else:
#       out = self._model.apply(self._params, model_inputs)

#     return ModelOutput(inputs=input_object, outputs=out)

#   def _serialise_params(self, fpath):
#     logger.info(f"Saving parameters to {fpath}")
#     raise NotImplementedError

#   def export(self, format: str, input_object: Any, export_model_path: str, **kwargs) -> ModelSpec:
#    raise NotImplementedError

#   @staticmethod
#   def deserialise(model_meta: ModelSpec) -> Tuple[Any, Any]:
#     raise NotImplementedError



"""
This is the hidden closet with all messy and cool things.

This submodule has files for managing the adapters for different frameworks. The important
repositories with the logic start with ``on_``, why?

#. ``on_ml``: contains the code for exporting and sharing the weights across different\
    ml libraries like pytorch, tensorflow, haiku, etc.
#. ``on_operators``: contains the code for exporting and importing operators from frameworks\
    like Airflow, Prefect, etc.
#. ``on_functions``: This library contains the code for static parsing of python and\
    creating the NBX-JobFlow

Then there is one main file for auto generating the code

#. ``autogen``: Which contains the compilers for different code generations like ``ml.py``,\
    there will be more in the future.

Then there are code that is generated:

#. ``ml``: This is the code for all ml frameworks, their exporting and importing using\
    message-stub format. That is all the user arguments became a dataclass which user has\
    to implement according to their requirments.
#. ``*_pb2.py/pyi``: These are the protobuf stubs generated, as of this writing there is only\
    one proto called ``ModelSpec`` which is like ``JobProto`` but for models.

The ``protos/`` folder contains all the proto definitions. Some of the files in here do not
have documentation and that is intentionally to hide all the complexities form the user.
If you are interested, you can read the source code directly from Github.

"""

# if you like what you see and want to work on more things like this, reach out research@nimblebox.ai

from .on_ml import  *
from .on_operators import *
from .model_spec_pb2 import *

try:
  from .ml import *
except ImportError:
  from .autogen import compile
  compile()
  from .ml import *


def get_model_functions(py_model):
  """Try to infer the functions from the model"""
  try:
    __import__(Framework_torch._load_framework)
    if Framework_torch._conditional(py_model):
      methods = Framework_torch._METHODS
      return methods
  except ImportError:
    pass
  return {}



"""
Network functions are gateway between NBX-Services. If you find yourself using this
you might want to reach out to us <research-at-nimblebox-dot-ai>!

But for the curious mind, many of our services work on gRPC and Protobufs. This network.py
manages the quirkyness of our backend and packs multiple steps as one function.
"""

import os
import requests
from time import sleep
from datetime import datetime, timedelta, timezone

import grpc

from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.field_mask_pb2 import FieldMask

from .hyperloop.nbox_ws_pb2 import UploadCodeRequest, CreateJobRequest, UpdateJobRequest
from .hyperloop.job_pb2 import Job as JobProto

from .init import nbox_grpc_stub
from .auth import secret
from .jobs import Job
from .messages import message_to_dict, rpc
from .framework.model_spec_pb2 import ModelSpec
from .subway import Sub30
from .utils import logger
from . import utils as U
from .hyperloop.job_pb2 import Job as JobProto


class NBXAPIError(Exception):
  pass


def deploy_serving(
  export_model_path,
  stub_all_depl: Sub30,
  model_spec: ModelSpec,
  wait_for_deployment=False,
):
  """One-Click-Deploy API to serve items on a NBX-Deploy

  Args:
    export_model_path (str): path to the file to upload
    stub_all_depl (nbox.Sub30): Subway RPC stub for ``/deployments``
    model_spec (nbox.ModelSpec): ModelSpec object
    wait_for_deployment (bool, optional): if true, acts like a blocking call (sync vs async)

  Returns:
    if ``wait_for_deployment == True`` returns ``(url, key)`` pair
  """
  logger.info(f"stub_all_depl: {stub_all_depl}")
  from nbox.auth import secret # it can refresh so add it in the method
  access_token = secret.get("access_token")
  URL = secret.get("nbx_url")

  # intialise the console logger
  logger.debug("-" * 30 + " NBX Deploy " + "-" * 30)
  logger.debug(f"Deploying on URL: {URL}")
  deployment_type = "nbox" # model_spec.deploy.type
  deployment_id = model_spec.deploy.id
  deployment_name = model_spec.deploy.name
  model_name = model_spec.name

  logger.debug(f"Deployment Type: '{deployment_type}', Deployment ID: '{deployment_id}'")

  if not deployment_id and not deployment_name:
    logger.debug("Deployment ID not passed will create a new deployment with name >>")
    deployment_name = U.get_random_name().replace("-", "_")

  file_size = os.stat(export_model_path).st_size // (1024 ** 2) # because in MB
  logger.debug(
    f"Deployment Name: '{deployment_name}', Model Name: '{model_name}', Model Path: '{export_model_path}', file_size: {file_size} MBs"
  )
  logger.debug("Getting bucket URL")

  # get bucket URL and upload the data
  out = stub_all_depl.u(deployment_id).get_upload_url(
    _method = "put",
    convert_args = "",
    deployment_meta = {},
    deployment_name = deployment_name,
    deployment_type = deployment_type, # "nbox" or "ovms2"
    file_size = str(file_size),
    file_type = "nbox",
    model_name = model_name,
    nbox_meta = message_to_dict(model_spec) , # message_to_json(model_spec), # annoying, but otherwise only the first key would be sent
    # deployment_id = deployment_id,
  )
  model_id = out["fields"]["x-amz-meta-model_id"]
  deployment_id = out["fields"]["x-amz-meta-deployment_id"]
  logger.debug(f"model_id: {model_id}")
  logger.debug(f"deployment_id: {deployment_id}")

  # upload the file to a S3 -> don't raise for status here
  logger.debug("Uploading model to S3 ...")
  r = requests.post(url=out["url"], data=out["fields"], files={"file": (out["fields"]["key"], open(export_model_path, "rb"))})

  # checking if file is successfully uploaded on S3 and tell webserver
  # whether upload is completed or not because client tells
  ws_stub_model = stub_all_depl.u(model_spec.deploy.id).models.u(model_id) # eager create the stub
  ws_stub_model.update(_method = "post", status = r.status_code == 204)

  # polling
  endpoint = None
  _stat_done = [] # status calls performed
  total_retries = 0 # number of hits it took
  access_key = None # this key is used for calling the model
  logger.debug(f"Check your deployment at {URL}/oneclick")
  if not wait_for_deployment:
    logger.debug("NBX Deploy")
    return endpoint, access_key

  logger.debug("Start Polling ...")
  while True:
    total_retries += 1

    # don't keep polling for very long, kill after sometime
    if total_retries > 50 and not wait_for_deployment:
      logger.debug(f"Stopping polling, please check status at: {URL}/oneclick")
      break

    sleep(5)

    # get the status update
    logger.debug(f"Getting updates ...")
    updates = ws_stub_model.history()["data"]
    for st in updates["model_history"]:
      curr_st = st["status"]
      logger.debug(f"Status: {curr_st}")
      if curr_st in _stat_done:
        continue
      logger.info(f"Status: {curr_st}")
      _stat_done.append(curr_st)

    if curr_st == "deployment.success":
      # if we do not have api key then query web server for it
      if access_key is None:
        server_endpoint = updates["model_data"]["api_url"]
        if server_endpoint is None:
          if wait_for_deployment:
            continue
          logger.debug("Deployment in progress ...")
          logger.debug(f"Endpoint to be setup, please check status at: {URL}/oneclick")
          break

    elif curr_st == "deployment.ready":
      out = stub_all_depl.u(model_spec.deploy.id).get_access_key()["data"]
      access_key = out["access_key"]

      # keep hitting /metadata and see if model is ready or not
      r = requests.get(url=f"{server_endpoint}/metadata", headers={"NBX-KEY": access_key, "Authorization": f"Bearer {access_token}"})
      if r.status_code == 200:
        logger.info(f"Model is ready")
        break

    # actual break condition happens here: bug in webserver where it does not return ready
    # curr_st == "ready"
    if access_key != None or "failed" in curr_st:
      break

  return server_endpoint, access_key

class Schedule:
  _days = {
    k:str(i) for i,k in enumerate(
      ["sun","mon","tue","wed","thu","fri","sat"]
    )
  }
  _months = {
    k:str(i+1) for i,k in enumerate(
      ["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"]
    )
  }

  def __init__(
    self,
    hour: int  = None,
    minute: int = None,
    days: list = [],
    months: list = [],
    starts: datetime = None,
    ends: datetime = None,
  ):
    """Make scheduling natural.

    Usage:

    .. code-block:: python

      # 4:20 everyday
      Schedule(4, 0)

      # 4:20 every friday
      Schedule(4, 20, ["fri"])

      # 4:20 every friday from jan to feb
      Schedule(4, 20, ["fri"], ["jan", "feb"])

      # 4:20 everyday starting in 2 days and runs for 3 days
      starts = datetime.now(timezone.utc) + timedelta(days = 2) # NOTE: that time is in UTC
      Schedule(4, 20, starts = starts, ends = starts + timedelta(days = 3))

      # Every 1 hour
      Schedule(1)

      # Every 69 minutes
      Schedule(minute = 69)

    Args:
        hour (int): Hour of the day, if only this value is passed it will run every ``hour``
        minute (int): Minute of the hour, if only this value is passed it will run every ``minute``
        days (list, optional): List of days (first three chars) of the week, if not passed it will run every day.
        months (list, optional): List of months (first three chars) of the year, if not passed it will run every month.
        starts (datetime, optional): UTC Start time of the schedule, if not passed it will start now.
        ends (datetime, optional): UTC End time of the schedule, if not passed it will end in 7 days.
    """
    self.hour = hour
    self.minute = minute

    self._is_repeating = self.hour or self.minute
    self.mode = None
    if self.hour == None and self.minute == None:
      raise ValueError("Atleast one of hour or minute should be passed")
    elif self.hour != None and self.minute != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      assert self.minute in list(range(0, 60)), f"Minute must be in range 0-59, got {self.minute}"
    elif self.hour != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.mode = "every"
      self.minute = "*"
      self.hour = f"*/{self.hour}"
    elif self.minute != None:
      self.hour = self.minute // 60
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.minute = f"*/{self.minute % 60}"
      self.hour = f"*/{self.hour}" if self.hour > 0 else "*"
      self.mode = "every"

    diff = set(days) - set(self._days.keys())
    if diff != set():
      raise ValueError(f"Invalid days: {diff}")
    self.days = ",".join([self._days[d] for d in days]) if days else "*"

    diff = set(months) - set(self._months.keys())
    if diff != set():
      raise ValueError(f"Invalid months: {diff}")
    self.months = ",".join([self._months[m] for m in months]) if months else "*"

    self.starts = starts or datetime.now(timezone.utc)
    self.ends = ends or datetime.now(timezone.utc) + timedelta(days = 7)

  @property
  def cron(self):
    """Cron string"""
    if self.mode == "every":
      return f"{self.minute} {self.hour} * * *"
    return f"{self.minute} {self.hour} * {self.months} {self.days}"

  def get_dict(self):
    return {
      "cron": self.cron,
      "mode": self.mode,
      "starts": self.starts,
      "ends": self.ends,
    }

  def get_message(self) -> JobProto.Schedule:
    """Get the JobProto.Schedule object for this Schedule"""
    _starts = Timestamp(); _starts.GetCurrentTime()
    _ends = Timestamp(); _ends.FromDatetime(self.ends)
    # return JobProto.Schedule(start = _starts, end = _ends, cron = self.cron)
    return JobProto.Schedule(cron = self.cron)

  def __repr__(self):
    return str(self.get_dict())


def deploy_job(zip_path: str, job_proto: JobProto):
  """Deploy an NBX-Job

  Args:
      zip_path (str): Path to the zip file
      schedule (Schedule): Schedule of the job
  Returns:
      nbox.Job: the job object
  """

  URL = secret.get("nbx_url")
  file_size = int(os.stat(zip_path).st_size // (1024 ** 2)) # in MBs

  # intialise the console logger
  logger.debug("-" * 30 + " NBX Jobs " + "-" * 30)
  logger.debug(f"Deploying on URL: {URL}")

  code = JobProto.Code(size = max(file_size, 1), type = JobProto.Code.Type.ZIP)
  job_proto.code.MergeFrom(code)

  # this means that this job already exists, we check if there are some metadata changes
  job_exists = job_proto.id != ""
  if job_exists:
    logger.debug("Found existing job, checking for update masks")
    old_job_proto = Job(job_proto.id, job_proto.auth_info.workspace_id).job_proto
    paths = []
    if old_job_proto.resource.SerializeToString(deterministic = True) != job_proto.resource.SerializeToString(deterministic = True):
      paths.append("resource")
    if old_job_proto.schedule.cron != job_proto.schedule.cron:
      paths.append("schedule.cron")

    # update the job
    if paths:
      logger.debug(f"Updating paths: {paths}")
      nbox_grpc_stub.UpdateJob(
        UpdateJobRequest(
          job = job_proto, update_mask = FieldMask(paths=paths)
        ),
      )

  # UploadJobCode is responsible for uploading the code of the job
  response: JobProto = rpc(
    nbox_grpc_stub.UploadJobCode,
    UploadCodeRequest(job = job_proto, auth = job_proto.auth_info),
    f"Failed to deploy job: {job_proto.id}"
  )
  job_proto.MergeFrom(response)

  s3_url = job_proto.code.s3_url
  s3_meta = job_proto.code.s3_meta
  logger.debug(f"Job ID: {job_proto.id}")

  logger.debug("Uploading model to S3 ...")
  r = requests.post(url=s3_url, data=s3_meta, files={"file": (s3_meta["key"], open(zip_path, "rb"))})
  try:
    r.raise_for_status()
  except:
    logger.error(f"Failed to upload model: {r.content.decode('utf-8')}")
    return

  logger.info("Creating new run ...")
  try:
    job: JobProto = nbox_grpc_stub.CreateJob(CreateJobRequest(job = job_proto))
  except grpc.RpcError as e:
    if e.code() == grpc.StatusCode.ALREADY_EXISTS:
      logger.debug(f"Job {job_proto.id} already exists")
    else:
      raise e
  except Exception as e:
    logger.error(f"Failed to create job: {e}")
    return

  # write out all the commands for this job
  logger.info("Run is now created, to 'trigger' programatically, use the following commands:")
  _api = f"nbox.Job(id = '{job_proto.id}', workspace_id='{job_proto.auth_info.workspace_id}').trigger()"
  _cli = f"python3 -m nbox jobs --id {job_proto.id} --workspace_id {job_proto.auth_info.workspace_id} trigger"
  _curl = f"curl -X POST {secret.get('nbx_url')}/api/v1/workpace/{job_proto.auth_info.workspace_id}/job/{job_proto.id}/trigger"
  _webpage = f"{secret.get('nbx_url')}/workspace/{job_proto.auth_info.workspace_id}/jobs/{job_proto.id}"
  logger.info(f" [python] - {_api}")
  logger.info(f"   [curl] - {_curl} -H 'Authorization: Bearer TOKEN'")
  logger.info(f"    [CLI] - {_cli}")
  logger.info(f"   [page] - {_webpage}")

  return Job(job_proto.id, job_proto.auth_info.workspace_id)



"""
``nbox.Job`` is a wrapper to the APIs that's it.

Notes
-----

* ``datetime.now(timezone.utc)`` is incorrect, use `this <https://blog.ganssle.io/articles/2019/11/utcnow.html>`_ method.
"""

import re
import sys
import os, re
import jinja2
import tabulate
from datetime import datetime, timezone
from google.protobuf.field_mask_pb2 import FieldMask

from . import utils as U
from .utils import logger
from .instance import Instance
from .version import __version__
from .init import nbox_grpc_stub
from .hyperloop.nbox_ws_pb2 import JobInfo
from .hyperloop.job_pb2 import NBXAuthInfo, Job as JobProto
from .hyperloop.nbox_ws_pb2 import ListJobsRequest, JobLogsRequest, ListJobsResponse, UpdateJobRequest
from .messages import message_to_dict, rpc, streaming_rpc


################################################################################
# NBX-Jobs Functions
# ==================
# These functions are assigned as static functions to the ``nbox.Job`` class.
################################################################################

def _repl_schedule(return_proto: bool = False):
  from .network import Schedule
  logger.info("Calendar Instructions (all time is in UTC Timezone):")
  logger.info("            What you want: Code")
  logger.info("\"   every 4:30 at friday\": Schedule(4, 30, ['fri'])")
  logger.info("\"every 12:00 on weekends\": Schedule(12, 0, ['sat', 'sun'])")
  logger.info("\"         every 10 hours\": Schedule(10)")
  logger.info("\"      every 420 minutes\": Schedule(minute = 420)")
  logger.info("> Enter the calender instruction: ")
  cron_instr = input("> ").strip()
  schedule_proto = None
  try:
    schedule_proto: Schedule = eval(cron_instr, {'Schedule': Schedule})
  except Exception as e:
    logger.error(f"Invalid calender instruction: {e}")
    sys.exit(1)
  if not return_proto:
    return cron_instr
  return schedule_proto

def _nbx_job(project_name: str):
  # Monday W34 [UTC 12 April, 2022 - 12:00:00]
  _ct = datetime.now(timezone.utc)
  _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
  created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"
  # created_time = None

  job_id_or_name = input("> Job ID or name: ")
  workspace_id = input("> Workspace ID (leave blank for personal): ")
  if not workspace_id:
    workspace_id = None

  scheduled = None
  logger.info("This job will run on NBX-Jobs")
  scheduled = input("> Is this a recurring job (y/N)? ").lower() == "y"
  cron_instr = None
  if scheduled:
    cron_instr = _repl_schedule()

  logger.info("This job will be scheduled to run on a recurring basis" if scheduled else "This job will run once")
  logger.info(f"Creating a folder: {project_name}")
  os.mkdir(project_name)
  os.chdir(project_name)
  py_data = dict(
    import_string_nbox = "from nbox.network import Schedule" if scheduled else None,
    job_id_or_name = job_id_or_name,
    workspace_id = workspace_id,
    scheduled = cron_instr,
    project_name = project_name,
    created_time = created_time,
  )
  py_f_data = {k:v for k,v in py_data.items() if v is not None}

  assets = U.join(U.folder(__file__), "assets")
  path = U.join(assets, "job_new.jinja")
  with open(path, "r") as f, open("nbx_user.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**py_f_data))

  path = U.join(assets, "job_nbx.jinja")
  with open(path, "r") as f, open("exe.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**py_f_data))

  md_data = dict(
    project_name = project_name,
    created_time = created_time,
    scheduled = scheduled,
  )

  path = U.join(assets, "job_new_readme.jinja")
  with open(path, "r") as f, open("README.md", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**md_data))

def _build_job(project_name):
  _ct = datetime.now(timezone.utc)
  _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
  created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"

  project_id = input("> Project ID: ")
  workspace_id = input("> Workspace ID (leave blank for personal): ")
  if not workspace_id:
    workspace_id = None

  inst = Instance(i = project_id, workspace_id = workspace_id)
  cpu, gpu_name, gpu_count = None, None, None
  if not inst.status == "RUNNING":
    logger.info("H/W Config Options:")
    logger.info("Since the instance is not running, you can select the hardware config for this job.")
    logger.info("there are two options (if no gpu is provided runs on cpu only):")
    logger.info("--cpu=2 [--gpu='<gpu_name>:<gpu_count>']")
    hw_config = input("> ").strip()
    splits = hw_config.split()
    if len(hw_config) < 3:
      logger.error(f"Invalid hardware config: {hw_config}")
      sys.exit(1)

    cpu = splits[0]
    gpu = None if len(splits) == 1 else splits[1]
    try:
      cpu = re.findall("^--cpu=(\d+)$", cpu)
      cpu = int(cpu)
      if gpu:
        gpu_name, gpu_count = re.findall("^--gpu='(.+):(\d+)'$", gpu)
        gpu_count = int(gpu_count)
    except:
      logger.error(f"Invalid hardware config: {hw_config}")
      sys.exit(1)
  
  py_data = dict(
    run_on_build = True,
    instance = None,
    import_string_others = "import subprocess",
    import_string_nbox = "from nbox.jobs import Instance",
    not_running = inst.status != "RUNNING",
    cpu_only = cpu and not gpu,
    cpu_count = cpu,
    gpu = gpu_name,
    gpu_count = gpu_count,
    workspace_id = workspace_id,
    project_name = inst.project_name,
    created_time = created_time,
  )
  py_f_data = {k:v for k,v in py_data.items() if v is not None}

  assets = U.join(U.folder(__file__), "assets")
  path = U.join(assets, "job_new.jinja")
  with open(path, "r") as f, open("nbx_user.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**py_f_data))

  path = U.join(assets, "job_nbx.jinja")
  with open(path, "r") as f, open("exe.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**py_f_data))

  md_data = dict(
    project_name = project_name,
    created_time = created_time,
  )

  path = U.join(assets, "job_new_readme.jinja")
  with open(path, "r") as f, open("README.md", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**md_data))

def new(project_name, b: bool = False):
  """Create a new folder, this can be run on NBX-Jobs or NBX-Deploy.

  Args:
    project_name (str): The name of the job folder
    b (bool, def. 'False'): If True, then job will run on an instance
    workspace_id (str, def. 'None'): If defined, that workspace will be used
      else personal workspace will be used
  """
  project_name = str(project_name)
  out = re.findall("^[a-zA-Z0-9_]+$", project_name)
  if not out:
    raise ValueError("Project name can only contain letters and underscore")

  if os.path.exists(project_name):
    raise ValueError(f"Project {project_name} already exists")

  fn = _build_job if b else _nbx_job
  fn(project_name)

  with open("requirements.txt", "w") as f:
    f.write(f"nbox=={__version__}")

  logger.debug("Completed")

def new_model(project_name):
  """Create a new folder, this can be run on NBX-Jobs or NBX-Deploy.

  Args:
    project_name (str): The name of the folder
  """
  project_name = str(project_name)
  out = re.findall("^[a-zA-Z0-9_]+$", project_name)
  if not out:
    raise ValueError("Folder name can only contain letters and underscore")

  if os.path.exists(project_name):
    raise ValueError(f"Folder {project_name} already exists")

  # Monday W34 [UTC 12 April, 2022 - 12:00:00]
  _ct = datetime.now(timezone.utc)
  _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
  created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"

  deployment_id_or_name = input("> Deployment ID or name: ")

  logger.info(f"Creating a folder: {project_name}")
  os.mkdir(project_name)
  os.chdir(project_name)
  py_data = dict(
    deployment_id_or_name = deployment_id_or_name,
    project_name = project_name,
    created_time = created_time,
    scheduled = None,
    import_string_nbox = None,
  )
  py_f_data = {k:v for k,v in py_data.items() if v is not None}

  assets = U.join(U.folder(__file__), "assets")
  path = U.join(assets, "job_new.jinja")
  with open(path, "r") as f, open("nbx_user.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**py_f_data))

  md_data = dict(
    project_name = project_name,
    created_time = created_time,
    scheduled = False,
  )

  path = U.join(assets, "job_new_readme.jinja")
  with open(path, "r") as f, open("README.md", "w") as f2:
    f2.write(jinja2.Template(f.read()).render(**md_data))

  with open("requirements.txt", "w") as f:
    f.write(f"nbox=={__version__}")

  logger.info("Process complete")


def get_job_list(workspace_id: str = None, sort: str = "name"):
  """Get list of jobs, optionally in a workspace"""
  auth_info = NBXAuthInfo(workspace_id = workspace_id)
  out: ListJobsResponse = rpc(
    nbox_grpc_stub.ListJobs,
    ListJobsRequest(auth_info = auth_info),
    "Could not get job list",
  )

  out = message_to_dict(out)
  if len(out["Jobs"]) == 0:
    logger.info("No jobs found")
    sys.exit(0)

  # filters = [f.upper() for f in filters]
  headers=list(out["Jobs"][0].keys())
  sorted_jobs = sorted(out["Jobs"], key = lambda x: x[sort])
  data = []
  for j in sorted_jobs:
    _row = []
    for x in headers:
      if x == "status":
        _row.append(JobProto.Status.keys()[j[x]])
        continue
      _row.append(j[x])
      # if "*" in filters:
      #   _row.append(j[x])
      # if j["status"] in filters:
      #   _row.append(j[x])
    data.append(_row)
  for l in tabulate.tabulate(data, headers).splitlines():
    logger.info(l)


################################################################################
# NimbleBox.ai Jobs
# =================
# This is the actual job object that users can manipulate. It is a shallow class
# around the NBX-Jobs gRPC API.
################################################################################

class Job:
  new = staticmethod(new)
  status = staticmethod(get_job_list)

  def __init__(self, id, workspace_id = None):
    """Python wrapper for NBX-Jobs gRPC API

    Args:
        id (str): job ID
        workspace_id (str, optional): If None personal workspace is used. Defaults to None.
    """
    self.id = id
    self.workspace_id = workspace_id
    self.job_proto = JobProto(id = id, auth_info = NBXAuthInfo(workspace_id = workspace_id))
    self.job_info = JobInfo(job=self.job_proto)
    self.refresh()

  def change_schedule(self, new_schedule: 'Schedule' = None):
    """Change schedule this job"""
    logger.debug(f"Updating job '{self.job_proto.id}'")
    if new_schedule == None:
      new_schedule = _repl_schedule(True) # get the information from REPL
    self.job_proto.schedule.MergeFrom(new_schedule.get_message())
    rpc(
      nbox_grpc_stub.UpdateJob,
      UpdateJobRequest(job=self.job_proto, update_mask=FieldMask(paths=["schedule"])),
      "Could not update job schedule",
      raise_on_error = True
    )
    logger.debug(f"Updated job '{self.job_proto.id}'")
    self.refresh()

  def __repr__(self) -> str:
    x = f"nbox.Job('{self.job_proto.id}', '{self.job_proto.auth_info.workspace_id}'): {self.status}"
    if self.job_proto.schedule.ByteSize != None:
      x += f" {self.job_proto.schedule}"
    else:
      x += " (no schedule)"
    return x

  def logs(self, f = sys.stdout):
    """Stream logs of the job, ``f`` can be anything has a ``.write/.flush`` methods"""
    logger.debug(f"Streaming logs of job '{self.job_proto.id}'")
    for job_log in streaming_rpc(
      nbox_grpc_stub.GetJobLogs,
      JobLogsRequest(job = JobInfo(job = self.job_proto)),
      f"Could not get logs of job {self.job_proto.id}, is your job complete?",
      True
    ):
      for log in job_log.log:
        f.write(log + "\n")
        f.flush()

  def delete(self):
    """Delete this job"""
    logger.info(f"Deleting job '{self.job_proto.id}'")
    rpc(nbox_grpc_stub.DeleteJob, JobInfo(job = self.job_proto,), "Could not delete job")
    logger.info(f"Deleted job '{self.job_proto.id}'")
    self.refresh()

  def refresh(self):
    """Refresh Job statistics"""
    logger.debug(f"Updating job '{self.job_proto.id}'")
    self.job_proto: JobProto = rpc(
      nbox_grpc_stub.GetJob, JobInfo(job = self.job_proto), f"Could not get job {self.job_proto.id}"
    )
    self.job_proto.auth_info.CopyFrom(NBXAuthInfo(workspace_id = self.workspace_id))
    self.job_info.CopyFrom(JobInfo(job = self.job_proto))
    logger.debug(f"Updated job '{self.job_proto.id}'")

    self.status = self.job_proto.Status.keys()[self.job_proto.status]

  def trigger(self):
    """Manually triger this job"""
    logger.info(f"Triggering job '{self.job_proto.id}'")
    rpc(nbox_grpc_stub.TriggerJob, JobInfo(job=self.job_proto), f"Could not trigger job '{self.job_proto.id}'")
    logger.debug(f"Triggered job '{self.job_proto.id}'")
    self.refresh()

  def __call__(self):
    return self.trigger()

  def pause(self):
    """Pause the execution of this job.
    
    WARNING: This will remove all the scheduled runs, if present""" 
    logger.info(f"Pausing job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.PAUSED
    rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not pause job {self.job_proto.id}", True)
    logger.debug(f"Paused job '{self.job_proto.id}'")
    self.refresh()
  
  def resume(self):
    """Resume the Job with the current schedule, if provided else simlpy sets status as ACTIVE"""
    logger.info(f"Resuming job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.SCHEDULED
    rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not resume job {self.job_proto.id}", True)
    logger.debug(f"Resumed job '{self.job_proto.id}'")
    self.refresh()



"""
Creates a socket tunnel between users localhost to server called RSockServer (Reverse Socket Server) .
Usage:
  client-tunnel.py <client_port>:<instance_name>:<instance_port> <auth>
Takes in the following arguments:
  - client_port: The port that the user can connect to.
  - instance_name: The name of the instance that the user wants to connect to.
  - instance_port: The port that the instance is listening on.
  - auth: The authentication token that the user has to provide to connect to the RSockServer.
  
"""

import os
import sys
import ssl
import socket
import socket
import threading
import subprocess
from time import sleep
from typing import List
from functools import partial
from datetime import datetime, timezone

from nbox.utils import NBOX_HOME_DIR, logger as nbx_logger
from nbox import utils as U
from nbox.auth import secret
from nbox.instance import Instance


class FileLogger:
  def __init__(self, filepath):
    self.filepath = filepath
    self.f = open(filepath, "a")

    self.debug = partial(self.log, level="debug",)
    self.info = partial(self.log, level="info",)
    self.warning = partial(self.log, level="warning",)
    self.error = partial(self.log, level="error",)
    self.critical = partial(self.log, level="critical",)

  def log(self, message, level):
    self.f.write(f"[{datetime.now(timezone.utc).isoformat()}] {level}: {message}\n")
    self.f.flush()


class RSockClient:
  """
  This is a RSockClient. It handels the client socket where client is the user application trying to connect to "client_port"
  Connects to RSockServer listening on localhost:886.
  RSockServer recieves instructions as a string and sends back a response.
  RSockServer requires following steps to setup
  First,
    Authentication:
      - Authentication happens by sending
        `"AUTH~{AUTH_TOKEN}"`
      - AUTH_TOKEN is not defined and is default to 'password'
    Seting config:
      - you can set config by sending
        `"SET_CONFIG~{instance}~{instance_port}"`
      - "instance" - Currently is the internal ip of the instance.
      - "instance_port" - What port users wants to connect to.
    Connect:
      - This Starts the main loop which
        1. Listen on client_port
        2. On connection, 
      2. On connection, 
        2. On connection, 
          a. Send AUTH
          b. If AUTH is successful, send SET_CONFIG
          c. If SET_CONFIG is successful, send CONNECT
          d. If CONNECT is successful, start io_copy
    IO_COPY:
      - This is the main loop that handles the data transfer between client and server. This is done by creating a new thread for each connection.
      - The thread is created by calling the function "io_copy" for each connection that is "server" and "client".
      - When a connection is closed, the loop is stopped.
  """

  def __init__(self, connection_id, client_socket, user, subdomain, instance_port, file_logger, auth, secure=False):
    """
    Initializes the client.
    Args:
      client_socket: The socket that the client is connected to.
      instance: The instance that the client wants to connect to.
      instance_port: The port that the instance is listening on.
      auth: The authentication token that the client has to provide to connect to the RSockServer.
      secure: Whether or not the client is using SSL.
    
    """
    self.connection_id = connection_id
    self.client_socket = client_socket
    self.user = user
    self.subdomain = subdomain
    self.instance_port = instance_port
    self.auth = auth
    self.secure = secure

    self.client_auth = False
    self.rsock_thread_running = False
    self.client_thread_running = False
    self.logger = file_logger

    self.log('Starting client')
    self.connect_to_rsock_server()
    self.log('Connected to RSockServer')
    # self.authenticate()
    # self.log('Authenticated client')
    self.set_config()
    self.log('Client init complete')

  def __repr__(self):
    return f"RSockClient({self.subdomain} | {self.connection_id})"
  
  def log(self, message, level="INFO"):
    self.logger.info(f"[{self.connection_id}] [{level}] {message}")
  
  def connect_to_rsock_server(self):
    """
    Connects to RSockServer.
    """
    self.log('Connecting to RSockServer', "DEBUG")
    rsock_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rsock_socket.connect(('rsock.rc.nimblebox.ai', 886))

    if self.secure:
      self.log('Starting SSL')
      certfile = U.join(U.folder(__file__), "pub.crt")
      self.rsock_socket = ssl.wrap_socket(rsock_socket, ca_certs=certfile, cert_reqs=ssl.CERT_REQUIRED)
    else:
      self.rsock_socket = rsock_socket

  def authenticate(self):
    """
    Authenticates the client.
    Sends `"AUTH~{AUTH_TOKEN}"` to RSockServer.
    """
    self.log('Authenticating client')
    self.rsock_socket.sendall(bytes('AUTH~{}'.format(self.auth), 'utf-8'))
    auth = self.rsock_socket.recv(1024)
    auth = auth.decode('utf-8')
    if auth == 'OK':
      self.log('Client authenticated')
    else:
      self.log('Client authentication failed', "ERROR")
      self.client_auth = False
      exit(1)
  
  def set_config(self):
    """
    Sets the config of the client.
    Sends `"SET_CONFIG~{instance}~{instance_port}"` to RSockServer.
    """
    self.log('Setting config')
    self.rsock_socket.sendall(bytes('SET_CLIENT~{}~{}~{}~{}'.format(self.user, self.subdomain, self.instance_port, self.auth), 'utf-8'))
    config = self.rsock_socket.recv(1024)
    config = config.decode('utf-8')
    self.log('Config set to {}'.format(config))
    if config == 'OK':
      self.client_auth = True
      self.log('Config set')
    else:
      self.log('Config set failed', "ERROR")
      exit(1)

  def connect(self):
    """
    Connects the client to RSockServer.
    Sends `"CONNECT"` to RSockServer.
    """
    if self.client_auth:
      self.log('Starting the io_copy loop')
      self.rsock_socket.sendall(bytes('CONNECT', 'utf-8'))

      connect_status = self.rsock_socket.recv(1024)
      connect_status = connect_status.decode('utf-8')
      self.log('Connected status: {}'.format(connect_status))
      if connect_status == 'OK':
        self.log('Connected to project...')
      else:
        self.log('Connect failed', "ERROR")
        exit(1)

      # start the io_copy loop
      self.rsock_thread_running = True
      self.client_thread_running = True

      self.rsock_thread = threading.Thread(target=self.io_copy, args=("server", ))
      self.client_thread = threading.Thread(target=self.io_copy, args=("client", ))

      self.rsock_thread.start()
      self.client_thread.start()
    else:
      self.log('Client authentication failed', "ERROR")
      exit(1)

  def io_copy(self, direction):
    """
    This is the main loop that handles the data transfer between client and server.
    """
    self.log('Starting {} io_copy'.format(direction))

    if direction == 'client':
      client_socket = self.client_socket
      server_socket = self.rsock_socket

    elif direction == 'server':
      client_socket = self.rsock_socket
      server_socket = self.client_socket

    while self.rsock_thread_running and self.client_thread_running:
      try:
        data = client_socket.recv(1024)
        if data:
          # self.log('{} data: {}'.format(direction, data))
          server_socket.sendall(data)
        else:
          self.log('{} connection closed'.format(direction))
          break
      except Exception as e:
        self.log('Error in {} io_copy: {}'.format(direction, e), "ERROR")
        break

    self.rsock_thread_running = False
    self.rsock_thread_running = False
    
    self.log('Stopping {} io_copy'.format(direction))
  
  def stop(self):
    """
    Stops the client.
    """
    self.log('Stopping client')
    self.rsock_thread_running = False
    self.client_thread_running = False
    self.client_socket.close()
    self.rsock_socket.close()
    self.log('Client stopped')


class ConnectionManager:
  def __init__(self, user: str, subdomain: str, file_logger: str, auth: str, notsecure: bool = False):
    """
    Args:
      localport: The port that the client will be listening on.
      user: The user that the client will be connecting as.
      subdomain: The subdomain that the client will be connecting to.
      port: The port that the server will be listening on.
      auth: The build auth token that the client will be using.
      notsecure: Whether or not to use SSL.
    """
    self.user = user
    self.subdomain = subdomain
    self.file_logger = file_logger
    self.auth = auth
    self.notsecure = notsecure
    self.connection_id = 0

    self.done = False
    self.clients = []
    self.threads = []
    self.connections = {}

  def __repr__(self) -> str:
    return f"ConnectionManager({self.subdomain}: {len(self.threads)} Connections)"

  def add(self, localport, buildport, *, _ssh: bool = True):
    """
    Adds a client to the list of clients.
    """
    t = threading.Thread(target = self.start, args = (localport, buildport, _ssh))
    t.start()
    self.threads.append(t)
    self.connections[localport] = buildport

  def start(self, localport, buildport, _ssh: bool = True):
    listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if not _ssh:
      listen_socket.settimeout(1.0)
    listen_socket.bind(('localhost', localport))
    listen_socket.listen(20)

    while not self.done:
      nbx_logger.debug("Waiting for connection on port {}".format(localport))
      try:
        client_socket, _ = listen_socket.accept()
      except socket.timeout:
        break

      nbx_logger.debug('Client connected')
      self.connection_id += 1
      nbx_logger.debug(f'Total clients connected -> {self.connection_id}')

      # create the client
      secure = not self.notsecure
      client = RSockClient(self.connection_id, client_socket, self.user, self.subdomain, buildport, self.file_logger, self.auth, secure)
      self.clients.append(client)

      # start the client
      client.connect()
    listen_socket.close()

  def quit(self):
    self.done = False
    sleep(2)
    for client in self.clients:
      client.rsock_socket.close()
    for thread in self.threads:
      thread.join()


def port_in_use(port: int) -> bool:
  with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    return s.connect_ex(('localhost', port)) == 0


def _create_threads(port: int, *apps_to_ports: List[str], i: str, workspace_id: str, _ssh: bool = True) -> ConnectionManager:
  def _sanity():
    if not (sys.platform.startswith("linux") or sys.platform == "darwin"):
      raise Exception(f"Unsupported platform '{sys.platform}', raise issue: https://github.com/NimbleBoxAI/nbox/issues")

    # sanity checks because python fire does not handle empty strings
    if i == "":
      raise Exception("Instance name cannot be empty")
    if workspace_id == "":
      raise Exception("Workspace ID cannot be empty")

  _sanity() # run sanity checks

  apps = {} # <localport-cloudport>
  for ap in apps_to_ports:
    try:
      localport, buildport = ap.split(':')
      apps[int(localport)] = int(buildport)
    except ValueError:
      raise ValueError(f"Incorrect local:build '{ap}', are you passing integers?")
  apps[port] = 2222 # hard code

  ports_used = []
  for k in apps:
    if port_in_use(k):
      ports_used.append(str(k))

  if ports_used:
    raise ValueError(f"Ports {', '.join(ports_used)} are already in use")

  # check if instance is the correct one
  instance = Instance(i, workspace_id)
  instance._unopened_error()

  # create logging for RSock
  folder = U.join(NBOX_HOME_DIR, "tunnel_logs")
  os.makedirs(folder, exist_ok=True)
  filepath = U.join(folder, f"tunnel_{instance.project_id}.log") # consistency with IDs instead of names
  file_logger = FileLogger(filepath)
  nbx_logger.info(f"Logging to {filepath}")

  # start the instance with _ssh mode
  instance.start(_ssh = True)

  # conman runs threads internally for multiple connections
  conman = ConnectionManager(
    file_logger = file_logger,
    user = secret.get("username"), 
    subdomain = instance.open_data.get("url"),
    auth = instance.open_data.get("token"),
  )
  for localport, buildport in apps.items():
    nbx_logger.debug(f"Creating connection from NBX:{buildport} -> local:{localport}")
    conman.add(localport, buildport, _ssh = _ssh)
  return conman


def tunnel(port: int, *apps_to_ports: List[str], i: str, workspace_id: str):
  """the nbox way to SSH into your instance.

  Usage:
    tunn.py 8000 -i "nbox-dev"

  Args:
    port: Local port for terminal
    *apps_to_ports: A tuple of values `buildport:localport`. For example, ``jupyter:8888`` or ``2001:8002``
    i(str): The instance to connect to
    pwd (str): password to connect to that instance.
  """
  
  connection = _create_threads(port, *apps_to_ports, i = i, workspace_id = workspace_id)

  try:
    # start the ssh connection on terminal    
    # there is more to know about passing shell = True, see:
    # https://medium.com/python-pandemonium/a-trap-of-shell-true-in-the-subprocess-module-6db7fc66cdfd
    # https://stackoverflow.com/questions/3172470/actual-meaning-of-shell-true-in-subprocess
    nbx_logger.info(f"Starting SSH ... for graceful exit press Ctrl+D then Ctrl+C")
    comm = "ssh"
    if U.ENVVARS.NBOX_SSH_NO_HOST_CHECKING(False):
      comm += " -o StrictHostKeychecking=no"
    comm += f" -p {port} ubuntu@localhost"
    subprocess.call(comm, shell=True)
  except KeyboardInterrupt:
    nbx_logger.info("KeyboardInterrupt, closing connections")
    connection.quit()

  connection.quit()
  sys.exit(0) # graceful exit


from nbox import Operator
from nbox.lib.shell import ShellCommand, Python

class YATW(Operator):
  # yet another time waster
  def __init__(self, s=3):
    super().__init__()
    self.s = s

  def forward(self, i = ""):
    import random
    from time import sleep

    y = self.s + random.random()
    print(f"Sleeping (YATW-{i}): {y}")
    sleep(y)


def init_function(s=0.1):
  # this random fuction sleeps for a while and then returns number
  from random import randint
  from time import sleep

  y = randint(4, 8) + s
  print(f"Sleeping (fn): {y}")
  sleep(y)
  return y


class Magic(Operator):
  # named after the Dr.Yellow train that maintains the Shinkansen
  def __init__(self):
    super().__init__()
    self.init = Python(init_function) # waste time and return something
    self.foot = YATW() # Oh, my what will happen here?
    self.cond1 = ShellCommand("echo 'Executing condition {cond}'")
    self.cond2 = ShellCommand("echo 'This is the {cond} option'")
    self.cond3 = ShellCommand("echo '{cond} times the charm :P'")

  def forward(self):
    t1 = self.init()
    self.foot(0)

    if t1 > 6:
      self.cond1(cond = 1)
    elif t1 > 10:
      self.cond2(cond = "second")
    else:
      self.cond3(cond = "Third")
